/*
    Copyright 2000-2011 Stefan Kuhr

    This file is part of procaff.

    procaff is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    procaff is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with procaff; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



    Copyright 2000-2011 Stefan Kuhr

    Diese Datei ist Teil von procaff

    procaff ist freie Software. Sie koennen procaff unter
    den Bedingungen der GNU General Public License, wie von der
    Free Software Foundation herausgegeben, weitergeben und/oder
    modifizieren, entweder unter Version 2 der Lizenz oder (wenn
    Sie es wuenschen) jeder spaeteren Version.

    Die Veroeffentlichung von procaff erfolgt in der
    Hoffnung, da es Ihnen von Nutzen sein wird, aber OHNE JEDE
    GEWAEHRLEISTUNG - sogar ohne die implizite Gewaehrleistung
    der MARKTREIFE oder der EIGNUNG FUER EINEN BESTIMMTEN ZWECK.
    Details finden Sie in der GNU General Public License.

    Sie sollten eine Kopie der GNU General Public License zusammen
    mit procaff erhalten haben. Falls nicht, schreiben Sie
    an die Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
    MA 02139, USA.

*/


// procaff.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include <malloc.h>
#define NOCRYPT //PSDK headers issue warnings if we don't exclude this crap
#pragma warning (disable: 4201 4214 4115)
#include <windows.h>
#pragma warning (default: 4201 4214 4115)
#include "procaffv.h"



void PrintProcessorUsage(DWORD_PTR dwpProcessAffinityMask);

BOOL WINAPI EnumProcs(LPCTSTR szProcessName, DWORD *lpdwPid, BOOL *lpdwFoundOnce, BOOL *lpdwFoundTwice);



void PrintOutErrorInfo(DWORD dwErr, LPCTSTR sz)
{
    LPVOID lpMsgBuf;
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
    );
    // Process any inserts in lpMsgBuf.
    // ...
    // Display the string.
    _tprintf( (LPCTSTR)_T("Error %ld in %s: \n%s\n"), dwErr, sz, lpMsgBuf );
    // Free the buffer.
    LocalFree( lpMsgBuf );
}


void PrintUsage(void)
{
    printf("\nprocaff.exe version %d.%d.%d.%d by Stefan Kuhr\nBuilt on %s at %s\n\n", MAJORUPPER,  MAJORLOWER , MINORUPPER , MINORLOWER, __DATE__, __TIME__);
    printf("Usage: procaff option affinity  cmdline/pid/process-name\n\n\n");
    printf("Where option can be one of the following:\n\n   /wait: start process synchronously\n   /run: start process asynchronously\n   /set: set affinity of already running process\n\n\n");
    
    printf("Examples:\n\nprocaff  /wait 2 notepad.exe starts notepad.exe on CPU 1 synchronously\n");
    printf("procaff  /run  1 notepad.exe starts notepad.exe on CPU 0 asynchronously\n");
    printf("procaff  /set  3 4296 sets processor affinity of process with pid 4296 to 3\n");
    printf("procaff  /set  3 foo.exe sets processor affinity of process \"foo.exe\"  to 3\n");


    printf("\n");
}


#define CSTR2STR(str) (LPTSTR)_tcscpy((LPTSTR)alloca(((_tcslen(str)+1)*sizeof(TCHAR))), str)


#ifdef __cplusplus
extern "C"
#endif
int _tmain(int argc, TCHAR* argv[])
{

    HANDLE hProcess;
    DWORD_PTR dwpProcessAffinityMask = 0L;
    int iProcessor=0;
    TCHAR *szCmdLine = GetCommandLine();
    TCHAR *szStart = NULL;
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    DWORD dwSize;
    LPTSTR szCommand;
    BOOL bNoWait = FALSE, bOpen=FALSE;
    BOOL bOldStyleCmdline = TRUE;



    if (argc<3)
    {
        PrintUsage();
    	return 0;
    }


    // skip the first cmdline parameter and keep in mind, it could be surrounded by double quotes (that is the else if condition)
    if  (!_tcsncmp( szCmdLine, argv[0],_tcslen(argv[0])) )
        szStart = & szCmdLine[_tcslen(argv[0])+1];
    else if ( _tcslen(szCmdLine)>(_tcslen(argv[0])+2) &&szCmdLine[0]=='\"' && szCmdLine[_tcslen(argv[0])+1]=='\"' && !_tcsncmp( &szCmdLine[1], argv[0],_tcslen(argv[0])))
        szStart = & szCmdLine[_tcslen(argv[0])+3];
    else
    {
        //bail! we have not been started in a standard style
        printf("Non standard process creation.\n");
        return 255;
    }


    if (argv[1][0] != _T('/') && argv[1][0] != _T('-') )
    {

        iProcessor = _ttoi(argv[1]);
        if (iProcessor <0)
        {
            PrintUsage();
    	    return 0;
        }


    }
    else
    {
        if (argc<4)
        {
            PrintUsage();
    	    return 0;
        }
        bOldStyleCmdline = FALSE;
        if (!_tcscmp(&argv[1][1], _T("wait")))
        {

        }
        else if (!_tcscmp(&argv[1][1], _T("run")))
        {
            bNoWait = TRUE;
        }
        else if (!_tcscmp(&argv[1][1], _T("set")))
        {
            bOpen = TRUE;
        }
        else
        {
            PrintUsage();
    	    return 0;
        }


        // now szStart may points on additional whitespace chars after the first cmdline parameter:
        while (_istspace(szStart[0]))// skip all the whitespaces
            szStart = _tcsinc(szStart);
    }


    //now do exactly the same for the second parameter:
    szCmdLine = szStart;
    if  (!_tcsncmp( szCmdLine, argv[1],_tcslen(argv[1])) )
        szStart = & szCmdLine[_tcslen(argv[1])+1];
    else if ( _tcslen(szCmdLine)>(_tcslen(argv[1])+2) &&szCmdLine[0]=='\"' && szCmdLine[_tcslen(argv[1])+1]=='\"' && !_tcsncmp( &szCmdLine[1], argv[1],_tcslen(argv[1])))
        szStart = & szCmdLine[_tcslen(argv[1])+3];
    else
    {
        //bail! we have not been started in a standard style
        printf("Non standard process creation.\n");
        return 255;
    }




    while (_istspace(szStart[0]))// skip all the whitespaces
        szStart = _tcsinc(szStart);


    if (!bOldStyleCmdline)
    {
        // if we have the new style command line, we have one additional command line parameter we have to skip:
        szCmdLine = szStart;
        if  (!_tcsncmp( szCmdLine, argv[2],_tcslen(argv[2])) )
            szStart = & szCmdLine[_tcslen(argv[2])+1];
        else if ( _tcslen(szCmdLine)>(_tcslen(argv[2])+2) &&szCmdLine[0]=='\"' && szCmdLine[_tcslen(argv[2])+1]=='\"' && !_tcsncmp( &szCmdLine[1], argv[2],_tcslen(argv[2])))
            szStart = & szCmdLine[_tcslen(argv[2])+3];
        else
        {
            //bail! we have not been started in a standard style
            printf("Non standard process creation.\n");
            return 255;
        }
        while (_istspace(szStart[0]))// skip all the whitespaces
            szStart = _tcsinc(szStart);
    }



    if (!bOpen)
    {
        // we start a process: szStart is the process to start
        hProcess = OpenProcess(PROCESS_SET_INFORMATION, FALSE, GetCurrentProcessId());
        if (!hProcess)
        {
            PrintOutErrorInfo(GetLastError(), (LPCTSTR)_T("OpenProcess"));
            return 255;
        }

        if (bOldStyleCmdline)
            dwpProcessAffinityMask = 1<<iProcessor;
        else
        {
            if(_T('0')==argv[2][0]  &&
               (_T('x')==argv[2][1] || _T('X')==argv[2][1]))
            {
                dwpProcessAffinityMask = (DWORD_PTR)_tcstoui64(&argv[2][2], NULL, 16);
            }    
            else
            {
                dwpProcessAffinityMask = (DWORD_PTR)_tcstoui64(argv[2], NULL, 10);
            }
        }

        if (!SetProcessAffinityMask(hProcess, dwpProcessAffinityMask))
        {
            CloseHandle(hProcess);
            PrintOutErrorInfo(GetLastError(),(LPCTSTR)_T("SetProcessAffinityMask"));
            return 255;
        }

        CloseHandle(hProcess);


        memset(&si,0, sizeof(si));
        si.cb = sizeof(STARTUPINFO);
        si.dwFlags = STARTF_USESHOWWINDOW;
        si.wShowWindow = (WORD) SW_SHOWNORMAL;
        memset(&pi,0, sizeof(pi));

        dwSize = ExpandEnvironmentStrings(szStart, NULL, 0L);
        if (!dwSize)
        {
            PrintOutErrorInfo(GetLastError(),(LPCTSTR)_T("ExpandEnvironmentStrings"));
            return 255;
        }

        szCommand = (LPTSTR)malloc(dwSize*sizeof(TCHAR));
        if (!szCommand)
        {
            PrintOutErrorInfo(ERROR_NOT_ENOUGH_MEMORY,(LPCTSTR)_T("malloc"));
            return 255;
        }

        ExpandEnvironmentStrings(szStart, szCommand, dwSize);

        if (CreateProcess(NULL, szCommand, NULL, NULL, FALSE, NORMAL_PRIORITY_CLASS, NULL, NULL, &si,&pi))
        {
            DWORD dwErr;
            if (bOldStyleCmdline)
                _tprintf((LPCTSTR)_T("Enabling only processor %d\n"), iProcessor);
            else
                PrintProcessorUsage(dwpProcessAffinityMask);
            
            if (!bNoWait)
            {
                _tprintf((LPCTSTR)_T("Process %s successfully started, waiting... \n"), szCommand);
                WaitForSingleObject(pi.hProcess, INFINITE);
                GetExitCodeProcess(pi.hProcess, &dwErr);
                _tprintf((LPCTSTR)_T("Process %s terminated with exit code %ld\n"), szCommand, dwErr);
            }
            else
                _tprintf((LPCTSTR)_T("Process %s successfully started.\n"), szCommand);
            CloseHandle(pi.hProcess);
            CloseHandle(pi.hThread);
            free(szCommand);
            return 0;
        }

        PrintOutErrorInfo(GetLastError(),(LPCTSTR)_T("CreateProcess"));
	    free(szCommand);

    }
    else
    {
        // we open a process: arv[4] is the pid of the process to change:
        DWORD dwProcID = (DWORD)_ttol(argv[3]);
        hProcess = OpenProcess(PROCESS_SET_INFORMATION, FALSE, dwProcID);
        if (!hProcess)
        {
            BOOL bOnce = FALSE, bTwice = FALSE;
            _tprintf(_T("Attempting to resolve %s as a process name...\n"), argv[3]);
            if (EnumProcs(argv[3], &dwProcID, &bOnce, &bTwice))
            {
                if (!bOnce)
                {
                    _tprintf(_T("Cannot find a process named %s or don't have the privileges to do so.\n"), argv[3]);
                    return 255;
                }                
                else if (bTwice)
                {
                    _tprintf(_T("There is more than one process named %s.\nPlease specify a Process ID instead.\n"), argv[3]);
                    return 255;
                }                
                else
                {
                    _tprintf(_T("Found Process Id %ld for %s\n"), dwProcID, argv[3]);
                    
                    hProcess = OpenProcess(PROCESS_SET_INFORMATION, FALSE, dwProcID);
                    if (!hProcess)
                    {
                        PrintOutErrorInfo(GetLastError(),(LPCTSTR)_T("OpenProcess"));
                        return 255;    
                    }
                }


            }
            else
            {
                _tprintf(_T("Can't load psapi.dll or psapi.dll does not export the required functions.\n"));
                return 255;
            }
        }

        if(_T('0')==argv[2][0]  &&
            (_T('x')==argv[2][1] || _T('X')==argv[2][1]))
        {
            dwpProcessAffinityMask = (DWORD_PTR)_tcstoui64(&argv[2][2], NULL, 16);
        }    
        else
        {
            dwpProcessAffinityMask = (DWORD_PTR)_tcstoui64(argv[2], NULL, 10);
        }
        
        if (!SetProcessAffinityMask(hProcess, dwpProcessAffinityMask))
        {
            CloseHandle(hProcess);
            PrintOutErrorInfo(GetLastError(),(LPCTSTR)_T("SetProcessAffinityMask"));
            return 255;
        }

        CloseHandle(hProcess);
        PrintProcessorUsage(dwpProcessAffinityMask);
    }
	return 0;
}


void PrintProcessorUsage(DWORD_PTR dwpProcessAffinityMask)
{
    DWORD dwCurrCPU=0;
    while (dwpProcessAffinityMask)
    {
        if (dwpProcessAffinityMask%2)
            _tprintf((LPCTSTR)_T("Enabling processor %ld\n"), dwCurrCPU);
        dwCurrCPU++;
        dwpProcessAffinityMask = dwpProcessAffinityMask>>1;
    }
}






/*
#include "EnumProc.h"
#include <tlhelp32.h>
#include <vdmdbg.h>
*/

/*
typedef struct
{
  DWORD          dwPID ;
  PROCENUMPROC   lpProc ;
  DWORD          lParam ;
  BOOL           bEnd ;
} EnumInfoStruct ;
*/

BOOL WINAPI EnumProcs(LPCTSTR szProcessName, DWORD *lpdwPid, BOOL *lpdwFoundOnce, BOOL *lpdwFoundTwice)
{
    HINSTANCE      hInstLib ;
    LPDWORD        lpdwPIDs ;
    DWORD          dwSize, dwSize2, dwIndex ;
    HMODULE        hMod ;
    HANDLE         hProcess ;
    TCHAR          szFileName[ MAX_PATH ] ;


    // PSAPI Function Pointers.
    BOOL (WINAPI *lpfEnumProcesses)( DWORD *, DWORD cb, DWORD * );
    BOOL (WINAPI *lpfEnumProcessModules)( HANDLE, HMODULE *,
     DWORD, LPDWORD );
    DWORD (WINAPI *lpfGetModuleFileNameEx)( HANDLE, HMODULE, LPTSTR, DWORD );





    // Load library and get the procedures explicitly. 
    hInstLib = LoadLibrary( _T("PSAPI.DLL") ) ;
    if( hInstLib == NULL )
    return FALSE ;

    // Get procedure addresses.
    lpfEnumProcesses = (BOOL(WINAPI *)(DWORD *,DWORD,DWORD*))    GetProcAddress( hInstLib, "EnumProcesses" ) ;
    lpfEnumProcessModules = (BOOL(WINAPI *)(HANDLE, HMODULE *,   DWORD, LPDWORD)) GetProcAddress( hInstLib, "EnumProcessModules" ) ;
    lpfGetModuleFileNameEx =(DWORD (WINAPI *)(HANDLE, HMODULE,    LPTSTR, DWORD )) GetProcAddress( hInstLib,    "GetModuleFileNameExW" ) ;
    if( lpfEnumProcesses == NULL ||    lpfEnumProcessModules == NULL ||    lpfGetModuleFileNameEx == NULL)
    {
       FreeLibrary( hInstLib ) ;
       return FALSE ;
    }

    // Call the PSAPI function EnumProcesses to get all of the
    // ProcID's currently in the system.
    // NOTE: In the documentation, the third parameter of
    // EnumProcesses is named cbNeeded, which implies that you
    // can call the function once to find out how much space to
    // allocate for a buffer and again to fill the buffer.
    // This is not the case. The cbNeeded parameter returns
    // the number of PIDs returned, so if your buffer size is
    // zero cbNeeded returns zero.
    // NOTE: The "HeapAlloc" loop here ensures that we
    // actually allocate a buffer large enough for all the
    // PIDs in the system.
    dwSize2 = 256 * sizeof( DWORD ) ;
    lpdwPIDs = NULL ;
    do
    {
        if( lpdwPIDs )
        {
           HeapFree( GetProcessHeap(), 0, lpdwPIDs ) ;
           dwSize2 *= 2 ;
        }
        lpdwPIDs = HeapAlloc( GetProcessHeap(), 0, dwSize2 );
        if( lpdwPIDs == NULL )
        {
            FreeLibrary( hInstLib ) ;
            return FALSE ;
        }
        if( !lpfEnumProcesses( lpdwPIDs, dwSize2, &dwSize ) )
        {
            HeapFree( GetProcessHeap(), 0, lpdwPIDs ) ;
            FreeLibrary( hInstLib ) ;
            return FALSE ;
        }
    }
    while( dwSize == dwSize2 ) ;

    // How many ProcID's did we get?
    dwSize /= sizeof( DWORD ) ;

    // Loop through each ProcID.
    for( dwIndex = 0 ; dwIndex < dwSize ; dwIndex++ )
    {
        szFileName[0] = 0 ;
        // Open the process (if we can... security does not
        // permit every process in the system).
        hProcess = OpenProcess(
           PROCESS_QUERY_INFORMATION | PROCESS_VM_READ,
           FALSE, lpdwPIDs[ dwIndex ] ) ;
        if( hProcess != NULL )
        {
            // Here we call EnumProcessModules to get only the
            // first module in the process this is important,
            // because this will be the .EXE module for which we
            // will retrieve the full path name in a second.
            if( lpfEnumProcessModules( hProcess, &hMod,sizeof( hMod ), &dwSize2 ) )
            {
              // Get Full pathname:
                if( !lpfGetModuleFileNameEx( hProcess, hMod,szFileName, sizeof( szFileName )/sizeof(TCHAR) ) )
                {
                    szFileName[0] = 0 ;
                }
                else
                {
                    TCHAR szFile[_MAX_FNAME];
                    TCHAR szExt[_MAX_EXT];
                    TCHAR szFileExt [_MAX_FNAME + _MAX_EXT - 1];
                    
                    _tsplitpath(szFileName, NULL, NULL, szFile, szExt);
                    _tcscpy(szFileExt, szFile);
                    _tcscat(szFileExt, szExt);

                    if (!_tcsicmp(szProcessName, szFileExt) || !_tcsicmp(szProcessName, szFile) )
                    {
                        *lpdwPid = lpdwPIDs[ dwIndex ];
                        if (*lpdwFoundOnce)
                            *lpdwFoundTwice = TRUE;
                        *lpdwFoundOnce = TRUE;
                    }


                }
            }
            CloseHandle( hProcess ) ;
        }
    
    }

    HeapFree( GetProcessHeap(), 0, lpdwPIDs ) ;

    // Free the library.
    FreeLibrary( hInstLib ) ;

    return TRUE ;
}

