// procaff.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include <malloc.h>
#include <windows.h>



void PrintOutErrorInfo(DWORD dwErr, LPCTSTR sz)
{
    LPVOID lpMsgBuf;
    FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM | 
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,
        NULL 
    );
    // Process any inserts in lpMsgBuf.
    // ...
    // Display the string.
    printf( "Error %ld in %s: \n%s\n", dwErr, sz, lpMsgBuf );
    // Free the buffer.
    LocalFree( lpMsgBuf );
}


void PrintUsage(void)
{
    printf(_T("procaff.exe by Stefan Kuhr, built on %s at %s\nUsage: procaff processor cmdline\nExample: procaff 1 notepad.exe will run notepad.exe on the 2nd processor\n\n"), __DATE__, __TIME__);
}


#define CSTR2STR(str) (LPTSTR)_tcscpy((LPTSTR)alloca(((_tcslen(str)+1)*sizeof(TCHAR))), str) 




int main(int argc, char* argv[])
{

    HANDLE hProcess = OpenProcess(PROCESS_SET_INFORMATION|PROCESS_QUERY_INFORMATION , FALSE, GetCurrentProcessId());

    if (!hProcess)
    {
        
        PrintOutErrorInfo(GetLastError(), _T("OpenProcess"));
        return 255;
    }
    DWORD dwProcessAffinityMask = 0L, dwSystemAffinityMask;
/*
    if (!GetProcessAffinityMask(hProcess, &dwProcessAffinityMask, &dwSystemAffinityMask))
    {
        
        PrintOutErrorInfo(GetLastError(),_T("GetProcessAffinityMask"));
        return 255;
    }
*/
    
    if (argc<3)
    {
        PrintUsage();
    	return 0;
    }
    
    int iProcessor = _ttoi(argv[1]);
    if (iProcessor <0)
    {
        PrintUsage();
    	return 0;
    }
    else
        printf(_T("Enabling only processor %d\n"), iProcessor);
    
    dwProcessAffinityMask = 1<<iProcessor;    
    if (!SetProcessAffinityMask(hProcess, dwProcessAffinityMask))
    {
        
        PrintOutErrorInfo(GetLastError(),_T("SetProcessAffinityMask"));
        return 255;
    }




    TCHAR *szCmdLine = GetCommandLine();
    TCHAR *szStart = NULL;


    // skip the first cmdline parameter and keep in mind, it could be surrounded by double quotes (that is the else if condition)
    if  (!_tcsncmp( szCmdLine, argv[0],_tcslen(argv[0])) )
        szStart = & szCmdLine[_tcslen(argv[0])+1];
    else if ( _tcslen(szCmdLine)>(_tcslen(argv[0])+2) &&szCmdLine[0]=='\"' && szCmdLine[_tcslen(argv[0])+1]=='\"' && !_tcsncmp( &szCmdLine[1], argv[0],_tcslen(argv[0])))
        szStart = & szCmdLine[_tcslen(argv[0])+3];
    else
    {
        //bail! we have not been started in a standard style
        printf("Non standard process creation.\n");
        return 255;
    }

    // now szStart may points on additional whitespace chars after the first cmdline parameter:
    while (_istspace(szStart[0]))// skip all the whitespaces
        szStart = _tcsinc(szStart);


    //now do exactly the same for the second parameter:
    szCmdLine = szStart;
    if  (!_tcsncmp( szCmdLine, argv[1],_tcslen(argv[1])) )
        szStart = & szCmdLine[_tcslen(argv[1])+1];
    else if ( _tcslen(szCmdLine)>(_tcslen(argv[1])+2) &&szCmdLine[0]=='\"' && szCmdLine[_tcslen(argv[1])+1]=='\"' && !_tcsncmp( &szCmdLine[1], argv[1],_tcslen(argv[1])))
        szStart = & szCmdLine[_tcslen(argv[1])+3];
    else
    {
        //bail! we have not been started in a standard style
        printf("Non standard process creation.\n");
        return 255;
    }




    while (_istspace(szStart[0]))// skip all the whitespaces
        szStart = _tcsinc(szStart);




    STARTUPINFO si;
    memset(&si,0, sizeof(si));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESHOWWINDOW;
    si.wShowWindow = (WORD) SW_SHOWNORMAL;
    PROCESS_INFORMATION pi;
    memset(&pi,0, sizeof(pi));
    
    if (CreateProcess(NULL, CSTR2STR(szStart), NULL, NULL, FALSE, NORMAL_PRIORITY_CLASS, NULL, NULL, &si,&pi))
    {
        printf(_T("Process %s successfully started, waiting... \n"), argv[2]);

        WaitForSingleObject(pi.hProcess, INFINITE);
        DWORD dwErr;
        GetExitCodeProcess(pi.hProcess, &dwErr);
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
        return dwErr;
    }  

    PrintOutErrorInfo(GetLastError(),_T("CreateProcess"));
	return 255;
}
